int dayLength = ...; //Working hours in each day
int weekLength = ...; //Working hours in each week
string dayNames[0..4] = ...;
range hourOfDay 1..dayLength; //Working hours in each day
enum dayOfWeek = ...; //Working days of the week
enum terms = ...; //Terms in the academic year
range StartTimes 0..weekLength-1;
//Weighting of hours of the day when Classes should be scheduled
int preferredTimes[StartTimes] = ...;

//Basic Types
enum CourseGroups = ...; //Group of courses whose classes must not overlap
//enum Courses = ...;
enum Modules = ...;
enum Classes = ...;
enum Lecturers = ...;
enum Rooms = ...;

//Relations
setof(Classes) ClassesInGroup[CourseGroups] = ...; //Classes in CourseGroup
setof(Modules) ModulesTaught[Lecturers] = ...; //Modules which can be taught by each lecturer
setof(Classes) ClassesInModule[Modules] = ...; //Classes in Modules
setof(Lecturers) LecturersForModule[Modules] = ...; //Lecturers available to teach module

//Attributes of Basic Types
int ClassSize[Classes] = ...;
int RoomSize[Rooms] = ...;

//Known and fixed entries on timetable
//Known entries are suggested solutions and may be changed; Fixed cannot be amended
int nrFixedClasses = ...; //Number of Classes whose position in timetable is fixed
int nrKnownClasses = ...; //Number of Classes whose possible position in timetable is known
range fixedClasses 1..nrFixedClasses;
range knownClasses 1..nrKnownClasses;

range Bool 0..1;

struct KeyedClasses {
   Classes class;
   StartTimes st;
   terms term;
   Rooms room;
   Lecturers lecturer;
};

KeyedClasses FixedClass[fixedClasses] = ...;
KeyedClasses KnownClass[knownClasses] = ...;


//Variables to solve for
var StartTimes StartTime[Classes];
var Rooms Location[Classes];
var Lecturers Lecturer[Classes];

var Bool softConstraints[knownClasses];

//maximize sum(i in knownClasses) softConstraints[i]
maximize sum(c in Classes) preferredTimes[StartTime[c]]

subject to {
   sum(i in knownClasses) softConstraints[i] = nrKnownClasses; //Use all KnownClasses
   forall (k in knownClasses) {
      softConstraints[k] = (
         (KnownClass[k].lecturer = Lecturer[KnownClass[k].class]) //Lecturer known
         & (KnownClass[k].room = Location[KnownClass[k].class]) //Room known
         & (KnownClass[k].st = StartTime[KnownClass[k].class])); //StartTime known
   };
   forall (f in fixedClasses) {
      (FixedClass[f].lecturer = Lecturer[FixedClass[f].class]); //Lecturer fixed
      (FixedClass[f].room = Location[FixedClass[f].class]); //Room fixed
      (FixedClass[f].st = StartTime[FixedClass[f].class]); //StartTime fixed
   };
   forall (m in Modules) 
      forall (c in ClassesInModule[m])
         (Lecturer[c] in LecturersForModule[m]); //Lecturer can teach Module

   forall (c in Classes)
      (RoomSize[Location[c]] >= ClassSize[c]); //Room must be big enough

   forall (m in Modules)
      forall (ordered c1, c2 in ClassesInModule[m]) {
         (Lecturer[c1] = Lecturer[c2]); //Only one lecturer for each module
      };

   forall (CG in CourseGroups) {
      alldifferent(all (c in ClassesInGroup[CG]) 
         StartTime[c]); //Avoid Clashes amoung classes in same CourseGroup
//      7 > sum(c in ClassesInGroup[CG]) (StartTime[c] < 9); //Max classes on Mon
//      7 > sum(c in ClassesInGroup[CG]) ((StartTime[c] > 8)&(StartTime[c] < 18)); //Max classes on Tue
//      7 > sum(c in ClassesInGroup[CG]) ((StartTime[c] > 17)&(StartTime[c] < 27)); //Max classes on Wed
//      7 > sum(c in ClassesInGroup[CG]) ((StartTime[c] > 26)&(StartTime[c] < 36)); //Max classes on Thur
//      7 > sum(c in ClassesInGroup[CG]) (StartTime[c] > 35); //Max classes on Fri
   };

   forall (ordered c1, c2 in Classes) {
      StartTime[c1] = StartTime[c2] => 
            Location[c1] <> Location[c2]; //Avoid Room clashes
      StartTime[c1] = StartTime[c2] => 
            Lecturer[c1] <> Lecturer[c2]; //Avoid Lecturer clashes
   };
   
//   forall (l in Lecturers)
//      19 > sum(c in Classes) (Lecturer[c] = l); //Max number of lectures given per week

};

search {
   forall(c in Classes) 
      once {
         tryall(i in StartTimes //Choose most preferred hours first
                ordered by decreasing preferredTimes[i])
            StartTime[c] = i;
         tryall(r in Rooms //Choose smallest adequate rooms first
                ordered by increasing RoomSize[r])
            Location[c] = r;
         tryall(l in Lecturers //Choose least assigned lecturers first
                ordered by increasing sum(c in Classes: bound(Lecturer[c])) (Lecturer[c] = l))
            Lecturer[c] = l;
   };
};

display(c in Classes) <"Module",c,StartTime[c],
                       "Day",dayNames[(StartTime[c]/dayLength)],
                       "Hour",(StartTime[c] mod dayLength) + 9,
                       "Room",Location[c],
                       "Lecturer",Lecturer[c]>;

